@@ -1,3 +1,4 @@
+from .exceptions import ConsumerUnhealthyException
 from .exceptions import SchemaConflictException
 from .exceptions import StopConsumer
 from .exceptions import UnhandledMessage
@@ -13,6 +14,7 @@ from .metrics import PUBLISHED_MESSAGES
 from .utils import resolve_dotted_name
 from asyncio.futures import Future
 from functools import partial
+from opentracing.scope_managers.contextvars import ContextVarsScopeManager
 from pydantic import BaseModel
 from pydantic import ValidationError
 from typing import Any
@@ -22,6 +24,7 @@ from typing import cast
 from typing import Dict
 from typing import List
 from typing import Optional
+from typing import Tuple
 from typing import Type
 
 import aiokafka
@@ -32,6 +35,7 @@ import asyncio
 import fnmatch
 import inspect
 import logging
+import opentracing
 import orjson
 import pydantic
 import signal
@@ -108,6 +112,8 @@ def _published_callback(topic: str, fut: Future) -> None:
 
 
 class SubscriptionConsumer:
+    _consumer: Optional[aiokafka.AIOKafkaConsumer] = None
+
     def __init__(
         self,
         app: "Application",
@@ -118,6 +124,10 @@ class SubscriptionConsumer:
         self._subscription = subscription
         self._event_handlers = event_handlers or {}
 
+    @property
+    def consumer(self) -> Optional[aiokafka.AIOKafkaConsumer]:
+        return self._consumer
+
     async def emit(self, name: str, *args: Any, **kwargs: Any) -> None:
         for func in self._event_handlers.get(name, []):
             try:
@@ -139,7 +149,7 @@ class SubscriptionConsumer:
             logger.info("Consumer stopped, exiting")
 
     async def __run(self) -> None:
-        consumer = aiokafka.AIOKafkaConsumer(
+        self._consumer = aiokafka.AIOKafkaConsumer(
             bootstrap_servers=cast(List[str], self._app._kafka_servers),
             loop=asyncio.get_event_loop(),
             group_id=self._subscription.group,
@@ -147,9 +157,11 @@ class SubscriptionConsumer:
             **self._app._kafka_settings or {},
         )
         pattern = fnmatch.translate(self._app.topic_mng.get_topic_id(self._subscription.stream_id))
-        listener = CustomConsumerRebalanceListener(consumer, self._app, self._subscription.group)
-        consumer.subscribe(pattern=pattern, listener=listener)
-        await consumer.start()
+        listener = CustomConsumerRebalanceListener(
+            self._consumer, self._app, self._subscription.group
+        )
+        self._consumer.subscribe(pattern=pattern, listener=listener)
+        await self._consumer.start()
 
         msg_handler = _raw_msg_handler
         sig = inspect.signature(self._subscription.func)
@@ -166,7 +178,19 @@ class SubscriptionConsumer:
         await self.emit("started", subscription_consumer=self)
         try:
             # Consume messages
-            async for record in consumer:
+            async for record in self._consumer:
+                tracer = opentracing.tracer
+                headers = {x[0]: x[1].decode() for x in record.headers or []}
+                parent = tracer.extract(opentracing.Format.TEXT_MAP, headers)
+                context = tracer.start_active_span(
+                    record.topic,
+                    tags={
+                        "message_bus.destination": record.topic,
+                        "message_bus.partition": record.partition,
+                        "message_bus.group_id": self._subscription.group,
+                    },
+                    references=[opentracing.follows_from(parent)],
+                )
                 CONSUMER_TOPIC_OFFSET.labels(
                     group_id=self._subscription.group,
                     stream_id=record.topic,
@@ -183,15 +207,19 @@ class SubscriptionConsumer:
                 try:
                     logger.info(f"Handling msg: {record}")
                     msg_data = orjson.loads(record.value)
-                    it = iter(sig.parameters.keys())
-                    kwargs: Dict[str, Any] = {next(it): msg_handler(record, msg_data)}
-                    for key in it:
+                    it = iter(sig.parameters.items())
+                    name, _ = next(it)
+                    kwargs: Dict[str, Any] = {name: msg_handler(record, msg_data)}
+
+                    for key, param in it:
                         if key == "schema":
                             kwargs["schema"] = msg_data["schema"]
                         elif key == "record":
                             kwargs["record"] = record
                         elif key == "app":
                             kwargs["app"] = self._app
+                        elif issubclass(param.annotation, opentracing.Span):
+                            kwargs[key] = context.span
 
                     with CONSUMED_MESSAGE_TIME.labels(
                         stream_id=record.topic,
@@ -219,14 +247,16 @@ class SubscriptionConsumer:
                     ).inc()
 
                 finally:
+                    context.span.finish()
+                    context.close()
                     await self.emit("message", record=record)
         finally:
             try:
-                await consumer.commit()
+                await self._consumer.commit()
             except Exception:
                 logger.info("Could not commit current offsets", exc_info=True)
             try:
-                await consumer.stop()
+                await self._consumer.stop()
             except Exception:
                 logger.warning("Could not properly stop consumer", exc_info=True)
 
@@ -317,12 +347,23 @@ class Application(Router):
         self._topic_prefix = topic_prefix
         self._replication_factor = replication_factor
         self._topic_mng: Optional[KafkaTopicManager] = None
+        self._subscription_consumers: List[SubscriptionConsumer] = []
 
     def mount(self, router: Router) -> None:
         self._subscriptions.extend(router.subscriptions)
         self._schemas.update(router.schemas)
         self._event_handlers.update(router.event_handlers)
 
+    async def health_check(self) -> None:
+        for subscription_consumer in self._subscription_consumers:
+            if (
+                subscription_consumer.consumer is not None
+                and not await subscription_consumer.consumer._client.ready(
+                    subscription_consumer.consumer._coordinator.coordinator_id
+                )
+            ):
+                raise ConsumerUnhealthyException(subscription_consumer, "Consumer is not ready")
+
     async def _call_event_handlers(self, name: str) -> None:
         handlers = self._event_handlers.get(name)
 
@@ -390,8 +431,24 @@ class Application(Router):
 
         producer = await self._get_producer()
 
+        tracer = opentracing.tracer
+        headers: Optional[List[Tuple[str, bytes]]] = None
+        if isinstance(tracer.scope_manager, ContextVarsScopeManager):
+            # This only makes sense if the context manager is asyncio aware
+            if tracer.active_span:
+                carrier: Dict[str, str] = {}
+                tracer.inject(
+                    span_context=tracer.active_span,
+                    format=opentracing.Format.TEXT_MAP,
+                    carrier=carrier,
+                )
+                headers = [(k, v.encode()) for k, v in carrier.items()]
+
         fut = await producer.send(
-            topic_id, value=orjson.dumps({"schema": schema_key, "data": data_}), key=key
+            topic_id,
+            value=orjson.dumps({"schema": schema_key, "data": data_}),
+            key=key,
+            headers=headers,
         )
 
         fut.add_done_callback(partial(_published_callback, topic_id))  # type: ignore
@@ -416,6 +473,7 @@ class Application(Router):
         return self._producer
 
     async def initialize(self) -> None:
+        await self._call_event_handlers("initialize")
 
         for reg in self._schemas.values():
             # initialize topics for known streams
@@ -456,10 +514,10 @@ class Application(Router):
         await self.finalize()
 
     async def consume_for(self, num_messages: int, *, seconds: Optional[int] = None) -> None:
-        consumers = []
 
         consumed = 0
 
+        self._subscription_consumers = []
         for subscription in self._subscriptions:
 
             async def on_message(record: aiokafka.structs.ConsumerRecord) -> None:
@@ -471,10 +529,10 @@ class Application(Router):
             consumer = SubscriptionConsumer(
                 self, subscription, event_handlers={"message": [on_message]}
             )
-            consumers.append(consumer)
+            self._subscription_consumers.append(consumer)
 
         try:
-            futures = [asyncio.create_task(c()) for c in consumers]
+            futures = [asyncio.create_task(c()) for c in self._subscription_consumers]
             future = asyncio.gather(*futures)
             if seconds is not None:
                 future = asyncio.wait_for(future, seconds)
@@ -490,14 +548,14 @@ class Application(Router):
                     fut.cancel()
 
     async def consume_forever(self) -> None:
-        consumers = []
+        self._subscription_consumers = []
 
         for subscription in self._subscriptions:
             consumer = SubscriptionConsumer(self, subscription)
-            consumers.append(consumer)
+            self._subscription_consumers.append(consumer)
 
         try:
-            futures = [asyncio.create_task(c()) for c in consumers]
+            futures = [asyncio.create_task(c()) for c in self._subscription_consumers]
             await asyncio.gather(*futures)
         finally:
             for fut in futures:
